/*----------------------------------------------------------------------
  File    : Edge.java
  Contents: management of the edges of a triangulation of a pointgon
  Author  : Christian Borgelt
  History : 2005.02.18 file created
            2005.02.25 function isects added
            2005.03.19 function distance added
            2006.02.17 second versions of distance and isects added
            2006.02.18 functions meets added, weight field added
            2013.04.22 adapted to type argument of Comparable
----------------------------------------------------------------------*/
package pointgon;

import java.io.Serializable;

/*--------------------------------------------------------------------*/
/** Class for the edges of a triangulation of a pointgon.
 *  @since  2005.02.18 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class Edge implements Comparable<Edge>, Serializable {

  private static final long serialVersionUID = 0x00010000L;

  /** edge identifier */
  protected int    id;
  /** source vertex */
  protected Vertex src;
  /** destination vertex */
  protected Vertex dst;
  /** weight/length of the edge */
  protected double wgt;

  /*------------------------------------------------------------------*/
  /** Create an edge.
   *  @param  src   the source vertex
   *  @param  dst   the destination vertex
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Edge (Vertex src, Vertex dst)
  { this.src = src; this.dst = dst; this.wgt = src.distance(dst); }

  /*------------------------------------------------------------------*/
  /** Create an edge.
   *  @param  src   the source vertex
   *  @param  dst   the destination vertex
   *  @param  wgt   the weight / length of the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Edge (Vertex src, Vertex dst, double wgt)
  { this.src = src; this.dst = dst; this.wgt = wgt; }

  /*------------------------------------------------------------------*/
  /** Get the length of the edge.
   *  @return the length of the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double length ()
  { return this.src.distance(this.dst); }

  /*------------------------------------------------------------------*/
  /** Check whether the edge meets a given edge.
   *  @param  edge  the edge to check
   *  @return whether the edge meets the given edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean meets (Edge edge)
  { return (this.src == edge.src) || (this.src == edge.dst)
        || (this.dst == edge.src) || (this.dst == edge.dst); }

  /*------------------------------------------------------------------*/
  /** Check whether the edge meets the given edge.
   *  @param  src   the source      vertex of the edge to check
   *  @param  dst   the destination vertex of the edge to check
   *  @return whether the edge meets the given edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean meets (Vertex src, Vertex dst)
  { return (this.src == src) || (this.src == dst)
        || (this.dst == src) || (this.dst == dst); }

  /*------------------------------------------------------------------*/
  /** Check whether the edge intersects a given edge.
   *  @param  edge  the edge to check
   *  @return whether the edge intersects the given edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isects (Edge edge)
  { return this.src.isectsX(this.dst, edge.src, edge.dst)
        && edge.src.isectsX(edge.dst, this.src, this.dst); }

  /*------------------------------------------------------------------*/
  /** Check whether the edge intersects the given edge.
   *  @param  src   the source      vertex of the edge to check
   *  @param  dst   the destination vertex of the edge to check
   *  @return whether the edge intersects the given edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isects (Vertex src, Vertex dst)
  { return this.src.isectsX(this.dst,      src,      dst)
        &&      src.isectsX(     dst, this.src, this.dst); }

  /*------------------------------------------------------------------*/
  /** Compute the distance of the edge to a given vertex.
   *  @param  v     the vertex to which to compute the distance
   *  @return the distance of the edge to a given vertex
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double distance (Vertex v)
  { return this.distance(v.x, v.y); }

  /*------------------------------------------------------------------*/
  /** Compute the distance of the edge to a given vertex.
   *  @param  x     the x-coordinate of the vertex
   *  @param  y     the y-coordinate of the vertex
   *  @return the distance of the edge to a given vertex
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double distance (double x, double y)
  {                            /* --- distance to given point */
    double sx, sy, dx, dy;     /* distances to source and destination */

    sx = this.src.x -x;        /* compute the coordinates of */
    sy = this.src.y -y;        /* the direction vectors to */
    dx = this.dst.x -x;        /* the source and the destination */
    dy = this.dst.y -y;        /* of the edge */
    if (sx *dx +sy *dy >= 0) { /* if distance to end is shortest */
      sx = sx*sx +sy*sy;       /* compute the squared distances */
      dx = dx*dx +dy*dy;       /* to the two end points */
      return Math.sqrt((sx < dx) ? sx : dx);
    }                          /* return the distance */
    dx = this.dst.x -this.src.x;
    dy = this.dst.y -this.src.y;
    x  = Math.sqrt(dx*dx +dy*dy);
    if (x <= 0) return Math.sqrt(sx*sx +sy*sy);
    return Math.abs(sy*dx -sx*dy) /x;
  }  /* distance() */          /* distance perpendicular to edge */

  /*------------------------------------------------------------------*/
  /** Compare two edges by their weight.
   *  @param  obj   the edge to compare to
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int compareTo (Edge obj)
  {                             /* --- compare two objects */
    if     (this.wgt > obj.wgt) return 1;
    return (this.wgt < obj.wgt) ? -1 : 0;
  }  /* compareTo() */

  /*------------------------------------------------------------------*/
  /** Create a string description of an edge.
   *  @return a string description of the edge
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public String toString ()
  { return "(" +this.src.id +", " +this.dst.id +")"; }

}  /* class Edge */
