/*----------------------------------------------------------------------
  File    : Path.java
  Contents: split a (sub-)pointgon with paths
  Author  : Christian Borgelt
  History : 2005.02.18 file created (as part of file MWT.java)
            2005.02.25 function valid added
            2005.03.01 functions split and next completed
            2005.03.02 functions inside and distribute added
            2005.03.04 functions nextHoles and valid debugged
            2005.03.05 subproblem key format changed, functions adapted
            2005.03.08 only triangles without holes cut off at anchor
            2005.03.09 changed to new processing scheme
            2005.03.15 some bugs in anchor assignments fixed
            2005.03.19 canonical representation of triangles
            2005.03.23 precomputed intersection table added
            2005.04.13 improved intersection test with bounding paths
            2005.06.20 bug in side tests in isect_ccw and isect_cw fixed
            2005.06.24 bug in split concerning type C -> type A fixed
            2005.12.15 use of more precomputed information added
            2005.12.28 changed to extension of abstract class
            2006.01.04 parameter mwt of init method moved to constructor
            2006.01.06 path creation routines reprogrammed
            2006.01.07 precheck of holes on edges exploited
            2006.01.08 incremental path check and path skipping added
            2006.01.19 anchor position search moved to init method
            2006.01.21 bug in function isects fixed (same line edges)
            2006.01.22 anchor removed from splitting paths (redundant)
            2006.01.26 forced perimeter anchor cut-off for two paths
            2006.02.05 adapted to make Combined1/2 subclasses
            2006.02.06 function distribute generalized
----------------------------------------------------------------------*/
package pointgon;

/*--------------------------------------------------------------------*/
/** Class for splitting a (sub-)pointgon with a path.
 *  @since  2006.01.27 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class Path extends Splitter {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00010000;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** fixed dummy path to represent any path cutting off an anchor */
  protected static final Vertex[] cut = new Vertex[0];

  /** first vertex on counterclockwise walk */
  protected Vertex   frst;
  /** last  vertex on counterclockwise walk */
  protected Vertex   last;
  /** indices of holes on current path */
  protected int[]    idxs;
  /** vertices  of current path */
  protected Vertex[] path;
  /** length    of current path */
  protected int      plen;
  /** direction of current path */
  protected int      pdir;
  /** position in holes for left  paths */
  protected int      lftpos;
  /** position in holes for right paths */
  protected int      rgtpos;
  /** full size of right subproblem */
  protected int      rgtall;
  /** full size of left  subproblem */
  protected int      lftall;
  /** length of the best path */
  protected int      bestlen;

  /*------------------------------------------------------------------*/
  /** Get hole positions for a path.
   *  @param  p    start vertex of the path
   *  @param  q    end   vertex of the path
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void getpos (Vertex p, Vertex q)
  {                             /* --- get hole positions for paths */
    int l, r, k;                /* indices for binary search */

    l = 0; r = holecnt;         /* do a binary search in the holes */
    while (l < r) {             /* for the relative position */
      k = (l+r) >> 1;           /* of the left end vertex */
      if (p.compareTo(holes[k]) > 0) l = k+1;
      else                           r = k;
    }                           /* holes with an index >= l are */
    this.lftpos = l;            /* right of the left end of the edge */
    if (q == p) {               /* if there is only one vertex, */
      this.rgtpos = l; return;} /* the second search can be skipped */
    r = holecnt;                /* do a binary search in the holes */
    while (l < r) {             /* for the relative position */
      k = (l+r) >> 1;           /* of the right end vertex */
      if (q.compareTo(holes[k]) > 0) l = k+1;
      else                           r = k;
    }                           /* holes with an index >= l are */
    this.rgtpos = l;            /* right of the right end of the edge */
  }  /* getpos() */

  /*------------------------------------------------------------------*/
  /** Get the next path end vertex.
   *  @return the next path end vertex.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Vertex nextEnd ()
  {                             /* --- get next path end vertex */
    int    i, n;                /* loop variable, number of vertices */
    Vertex p, q;                /* path vertices */

    n = this.verts.length;      /* get modulo for id computations */
    i = this.state;             /* get the current path state */
    if (i < 0) {                /* if end is in perimeter section */
      q = this.verts[(this.path[this.plen-1].id +n +this.dir) % n];
      p = this.key[-i % this.cnt];    /* get the next end vertex */
      if (p == q) {             /* if at end of perimeter section, */
        this.state = 1-i;       /* switch back to normal mode */
        if (q == this.key[0])   /* if already back at the anchor, */
          return null;          /* there are no more end vertices */
      } }
    else {                      /* if on one of the bounding paths */
      if (i >= this.cnt)        /* if at the end of the key, */
        return null;            /* there are no more end vertices */
      q = this.key[this.state++];  /* get the next key vertex */
      if (q == null) {          /* if at the perimeter section, */
        q = this.verts[(this.key[i-1].id +n +this.dir) % n];
        this.state = -1-i;      /* get the next perimeter vertex */
      }                         /* and note in the state that the */
    }                           /* perimeter section is processed */
    return this.path[this.plen-1] = q;
  }  /* nextEnd() */            /* store and return next end vertex */

  /*------------------------------------------------------------------*/
  /** Get the next holes sequence.
   *  @param  valid  whether preceding sequence was valid
   *  @return whether there is a next hole sequence
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean nextHoles (boolean valid)
  {                             /* --- get next holes sequence */
    int i;                      /* indices of the hole vertices */

    if (this.pdir > 0) {        /* if in left to right paths */
      i = this.idxs[this.plen-2];
      if (++i < this.holecnt) { /* if not at rightmost hole */
        if (!valid)             /* if prec. sequence was invalid, */
          this.plen--;          /* do not add, but replace last hole */
        this.idxs[this.plen-1] = i;
        this.path[this.plen-1] = this.holes[i];
        this.plen++;            /* add the next hole to the right */
        return true;            /* to the path, increase the length, */
      }                         /* and return 'next holes collected' */
      if (--this.plen > 1) {    /* if a hole is left on the path */
        i = ++this.idxs[this.plen-2];
        this.path[this.plen-2] = this.holes[i];
        return true;            /* go to the next hole to the right */
      }                         /* (which must exist in this case) */
      return false;             /* if there is no hole left, */
    }                           /* all holes sequences were generated */
    if (this.pdir < 0) {        /* if in right to left paths */
      i = this.idxs[this.plen-2];
      if (--i >= 0) {           /* if not at leftmost hole */
        if (!valid)             /* if prec. sequence was invalid, */
          this.plen--;          /* do not add, but replace last hole */
        this.idxs[this.plen-1] = i;
        this.path[this.plen-1] = this.holes[i];
        this.plen++;            /* add the next hole to the left */
        return true;            /* to the path, increase the length, */
      }                         /* and return 'next holes collected' */
      if (--this.plen > 1) {    /* if a hole is left on the path */
        i = --this.idxs[this.plen-2];
        this.path[this.plen-2] = this.holes[i];
        return true;            /* go to the next hole to the left */
      }                         /* (which must exist in this case) */
    }
    if ((this.pdir   >= 0)      /* if there are unprocessed holes */
    &&  (this.lftpos >  0)) {   /* to the left of the anchor */
      this.path[0] = this.holes[this.idxs[0] = this.lftpos-1];
      this.plen    =  2;        /* create first right to left path */
      this.pdir    = -1;        /* (starting with a single hole) */
      return true;              /* set the path direction and */
    }                           /* return 'next holes collected' */
    if (this.rgtpos < this.holecnt) {
      this.path[0] = this.holes[this.idxs[0] = this.rgtpos];
      this.plen    =  2;        /* create first left to right path */
      this.pdir    = +1;        /* (starting with a single hole) */
      return true;              /* set the path direction and */
    }                           /* return 'next holes collected' */
    return false;               /* there are no more hole sequences */
  }  /* nextHoles() */

  /*------------------------------------------------------------------*/
  /** Check the path cutting off the anchor.
   *  @param  p    the first vertex of the path
   *  @param  q    the last  vertex of the path
   *  @return whether the cut is valid
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean validcut (Vertex p, Vertex q)
  {                             /* --- check path cutting off anchor */
    int    i, n;                /* loop variable, number of vertices */
    Vertex v, h;                /* to traverse the vertices/holes */
    Vertex cw, ccw;             /* (counter)clockwise vertex */

    if ((this.key[1] != null) && (this.key[2] != null))
      ccw = this.key[2];        /* get third perimeter vertex */
    else {                      /* if vertex is in perimeter section */
      n = this.verts.length;    /* get modulo for id computations */
      ccw = this.verts[(p.id +n +this.dir) % n];
    }                           /* get third perimeter vertex */
    v = this.key[0];            /* get the anchor of the subproblem */
    if (!  v.isLeftOf (p, q)    /* if the anchor is not to the left */
    || ( ccw.isLeftOf (v, p)    /* or the next perimeter vertex is */
    &&  !ccw.isRightOf(p, q)))  /* inside the cut off triangle, */
      return false;             /* the path is invalid */
    /* An edge that cuts off the anchor (edge from the vertex that  */
    /* is counterclockwise from the anchor to the vertex that is    */
    /* clockwise from the anchor) must have the anchor to its left  */
    /* to be valid. In addition, the third vertex on the perimeter  */
    /* of the subproblem, provided it is to the left of the edge    */
    /* to the vertex counterclockwise from the anchor, must be to   */
    /* its right. Otherwise the edge is not inside the subproblem.  */
    if (this.trtab != null) {   /* if there is precomputed info. */
      if (this.trtab[v.id][p.id][q.id] == 0)
        return false; }         /* the triangle must be valid */
    else {                      /* if no precomputed info. available */
      for (i = this.holecnt; --i >= 0; )
        if (this.holes[i].isInside(v, p, q))
          return false;         /* if a hole is inside the triangle, */
    }                           /* the path is invalid */
    /* If the triangle that is cut off by the path contains a hole, */
    /* there must be a path starting at the anchor and leading      */
    /* to a hole vertex. Hence we need not consider the path that   */
    /* cuts off the anchor as a special case, the paths suffice.    */
    /* Note that the above loop need not check whether a hole is    */
    /* equal to a triangle corner before it checks whether the hole */
    /* lies inside the triangle, as all triangle corners lie on the */
    /* perimeter of the current subproblem.                         */
    return true;                /* return 'cut is valid' */
  }  /* validcut() */

  /*------------------------------------------------------------------*/
  /** Check whether path leads inside.
   *  @param  p    the first vertex of the path
   *  @param  q    the last  vertex of the path
   *  @return whether the path leads inside
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean inside (Vertex p, Vertex q)
  {                             /* --- check if path leads inside */
    int     n;                  /* number of vertices */
    Vertex  cw, ccw;            /* (counter)clockwise vertex */
    boolean rgt, lft;           /* flags for point positions */

    n   = this.verts.length;    /* get modulo for id computations */
    ccw = this.key[1];          /* and the ccw perimeter vertex */
    ccw = (ccw != null) ? ccw : this.verts[(p.id +n +this.dir) % n];
    lft = q.isLeftOf(p, ccw);   /* check path vertex against perim. */
    cw  = this.key[this.cnt-1]; /* get clockwise perimeter vertex */
    cw  = (cw  != null) ? cw  : this.verts[(p.id +n -this.dir) % n];
    rgt = q.isRightOf(p, cw);   /* check path vertex against perim. */
    if (p.isLeftOf(ccw, cw)) {  /* if pointgon is convex  at anchor */
      if (!lft || !rgt) return false; }
    else {                      /* if pointgon is concave at anchor */
      if (!lft && !rgt) return false; }
    /* A path starting at the anchor must be in a direction leading */
    /* to the inside of the subproblem pointgon. That is, it must   */
    /* lie "between" the two edges that are incident to the anchor. */
    return true;                /* return 'path leads inside' */
  }  /* inside() */

  /*------------------------------------------------------------------*/
  /** Check the added path edge.
   *  @return whether the added path edge is valid
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean valid ()
  {                             /* --- check added path edge */
    Vertex p, q;                /* end vertices of added edge */

    if (this.plen > 2) {        /* if there is more than one hole */
      p = this.path[this.plen-3];  /* get and check the edge that */
      q = this.path[this.plen-2];  /* was added to the sequence */
      return (this.wgts[p.id][q.id] >= 0)
          && !this.isects(p, q);
    }                           /* return whether edge is valid */
    p = this.key[0];            /* get and check the edge */
    q = this.path[0];           /* from the anchor and the path */
    return (this.wgts[p.id][q.id] >= 0)
        &&  this.inside(p, q)   /* return whether edge is valid */
        && !this.isects(p, q);  /* (leads inside and no intersection) */
  }  /* valid() */

  /*------------------------------------------------------------------*/
  /** Check whether a hole is inside a subproblem.
   *  @param  hole  the hole to check
   *  @param  key   the key  of the subproblem
   *  @param  cnt   the size of the subproblem
   *  @return whether the hole is inside the subproblem
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean inside (Vertex hole, Vertex[] key, int cnt)
  {                             /* --- check whether a hole is inside */
    int    i, r, n, c = 0;      /* loop variables, buffer, counter */
    Vertex p, q, v;             /* perimeter vertices of subproblem */

    n = this.verts.length;      /* get modulo for id computations */
    p = v = key[0];             /* traverse vertices on the perimeter */
    for (i = 1; i <= cnt; p = q) {         /* of the given subproblem */
      if (i < 0) {              /* if in the perimeter section */
        q = this.verts[(p.id +n +this.dir) % n];
        if (q == v) i = 1-i; }  /* check whether at end of section */
      else {                    /* if on one of the x-monotone paths */
        q = key[i++ % cnt];     /* get the next key vertex */
        if (q == null) {        /* if at the perimeter section */
          v = key[i % cnt];     /* get the next key element */
          q = this.verts[(p.id +n +this.dir) % n];
          i = -i;               /* get the next perimeter vertex */
        }                       /* and note in the index that the */
      }                         /* perimeter section is processed */
      /* See method "nextEnd" for more explanations of the above. */
      r = hole.horzLineCuts(p, q); /* check whether horz. line cuts */
      if (r == 0) return false; /* if vertex is on the edge, abort */
      if (r >  0) c++;          /* if horz. line cuts to the right */
    }                           /* of the vertex, count the edge */
    /* The idea of the above check is to shoot a straight line from   */
    /* the given point to the right and to count how many times this  */
    /* line cuts the perimeter. If the number is even, the point lies */
    /* outside the perimeter, if it is odd, the point lies inside.    */
    return (c & 1) != 0;        /* check the number of intersections */
  }  /* inside() */             /* (odd: inside, even: outside) */

  /*------------------------------------------------------------------*/
  /** Get the full size of a subproblem.
   *  @param  key   the key of the subproblem
   *  @param  cnt   the size of the subproblem
   *  @return the full size of a subproblem
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected int size (Vertex[] key, int cnt)
  { return (key == this.lftkey) ? this.lftall : this.rgtall; }

  /*------------------------------------------------------------------*/
  /** Create a path splitter.
   *  @param  mwt   the minimum weight triangulation object
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Path (MWT mwt)
  {                             /* --- create a subproblem splitter */
    int n = mwt.holes.length;   /* number of holes */
    this.dist     = ((mwt.mode & MWT.DISTHOLES) != 0);
    this.verts    = mwt.verts;  /* store perimeter vertices */
    this.dir      = mwt.dir;    /* and their order */
    this.wgts     = mwt.wgts;   /* note the precomputed information */
    this.trtab    = mwt.trtab;  /* and create the vertex vectors */
    this.rgtkey   = new Vertex[n +4];
    this.lftkey   = new Vertex[n +4];
    this.rgtholes = new Vertex[n];
    this.lftholes = new Vertex[n];
    this.idxs     = new int   [(n > 0) ? n : 1];
    this.path     = new Vertex[n +2];
    this.best     = new Vertex[n +1];
  }  /* Path() */

  /*------------------------------------------------------------------*/
  /** Initialize the path splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  holes    the holes of the subproblem
   *  @param  holecnt  the number of holes of the subproblem
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void init (Vertex[] key,   int cnt,
                       Vertex[] holes, int holecnt)
  {                             /* --- init. a subproblem splitter */
    int    n;                   /* number of perimeter vertices */
    Vertex v, p, q;             /* subproblem anchor, adjacent verts. */

    this.key     = key;         /* note the key */
    this.cnt     = cnt;         /* of the subproblem */
    this.holes   = holes;       /* note the holes that are (possibly) */
    this.holecnt = holecnt;     /* inside this subproblem */
    this.pdir    = 0;           /* init. the split state */
    this.state   = 0;           /* (no splits created yet) */
    this.bestlen = 0;           /* init. the best path length */
    n = this.verts.length;      /* get modulo for id computations */
    v = key[0];                 /* get the subproblem anchor */
    p = key[1];                 /* and the key elements that are */
    q = key[cnt-1];             /* adjacent to anchor (2nd and last) */
    this.frst = (p != null) ? p : this.verts[(v.id +n +this.dir) % n];
    this.last = (q != null) ? q : this.verts[(v.id +n -this.dir) % n];
    if (holecnt <= 0) this.lftpos = this.rgtpos = 0;
    else this.getpos(v, v);     /* get hole positions for paths */
  }  /* init() */

  /*------------------------------------------------------------------*/
  /** Initialize a path splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  spobj    the splitter object
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void init (Vertex[] key, int cnt, Object spobj)
  {                             /* --- init. a subproblem splitter */
    int    n;                   /* number of perimeter vertices */
    Vertex v, p, q;             /* subproblem anchor, adjacent verts. */

    this.key = key;             /* note the key */
    this.cnt = cnt;             /* of the subproblem */
    if (spobj instanceof Vertex) {
      this.plen    = 1;         /* if split object is a single edge, */
      this.pdir    = 0;         /* get the end vertex of this edge */
      this.path[0] = (Vertex)spobj; }
    else {                      /* if the split object is a path */
      this.plen = ((Vertex[])spobj).length;
      if (this.plen <= 0)       /* if the path is empty, */
        this.pdir = 0;          /* only set the path direction */
      else {                    /* if the path is not empty */
        System.arraycopy(spobj, 0, this.path, 0, this.plen);
        this.pdir = this.path[0].compareTo(key[0]);
      }                         /* copy the splitting path and */
    }                           /* determine the path direction */
    n = this.verts.length;      /* get modulo for id computations */
    v = key[0];                 /* get the subproblem anchor */
    p = key[1];                 /* and the key elements that are */
    q = key[cnt-1];             /* adjacent to anchor (2nd and last) */
    this.frst = (p != null) ? p : this.verts[(v.id +n +this.dir) % n];
    this.last = (q != null) ? q : this.verts[(v.id +n -this.dir) % n];
  }  /* init() */               /* note first and last key vertex */

  /*------------------------------------------------------------------*/
  /** Create the next subproblem split.
   *  @return whether a new split was created
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean next ()
  {                             /* --- create next subproblem split */
    int     n;                  /* number of perimeter vertices */
    Vertex  p, q;               /* end points of path */
    boolean valid;              /* whether prec. sequence was valid */

    /* --- path that cuts off the anchor --- */
    if (this.state == 0) {      /* if no path has been created yet */
      this.plen  = 0;           /* create path cutting off anchor and */
      this.state = 1;           /* switch to paths starting at anchor */
      p = this.frst;            /* get the two path vertices, i.e., */
      q = this.last;            /* vertices adjacent to the anchor */
      if ((this.wgts[p.id][q.id] >= 0)
      &&   this.validcut(p, q)  /* if the only path edge is valid, */
      &&  !this.isects(p, q))   /* the cut is valid (empty triangle), */
        return true;            /* and edge does not intersect paths, */
      this.plen = -1;           /* return 'next split created' */
    }                           /* invalidate the path length */

    /* --- single edge path starting at the anchor --- */
    if (this.pdir == 0) {       /* if no holes processed yet */
      this.plen = 1;            /* path contains only the end vertex */
      p = this.key[0];          /* get the subproblem anchor */
      while (true) {            /* end vertex search loop */
        q = this.nextEnd();     /* get the next end of the path */
        if (q == null) break;   /* if no more end vertices, abort */
        if ((q != this.frst)    /* edges to vertices adjacent */
        &&  (q != this.last)    /* to the anchor are not allowed */
        &&  (this.wgts[p.id][q.id] >= 0)
        &&   this.inside(p, q)  /* if the edge of the path is valid, */
        &&  !this.isects(p, q)) /* leads to the inside of pointgon, */
          return true;          /* and does not intersect a path, */
      }                         /* return 'next split created' */
      if (this.holecnt <= 0)    /* if there are no holes, */
        return false;           /* all splits have been generated */
      do {                      /* hole sequence generation loop */
        if (!this.nextHoles(false))
          return false;         /* check for another hole sequence */
      } while (!this.valid());  /* while hole sequence is invalid */
      this.state = 1;           /* reinitialize the path state */
    }                           /* for the end vertex traversal */

    /* --- multi-edge path starting at the anchor --- */
    while (true) {              /* path generation loop */
      q = this.nextEnd();       /* get the next end vertex */
      if (q != null) {          /* if there is another end vertex */
        p = this.path[this.plen-2];  /* get preceding path vertex */
        if ((this.wgts[p.id][q.id] < 0)
        ||  (q.compareTo(p) != this.pdir)
        ||  this.isects(p, q))  /* if the edge is invalid, has wrong */
          continue;             /* dir, or intersects paths, skip it */
        return true;            /* return 'next split created' */
      }                         /* (multi-edge path from anchor) */
      valid = true;             /* preceding hole sequence was valid */
      do {                      /* hole sequence generation loop */
        if (!this.nextHoles(valid))
          return false;         /* check for another hole sequence */
        valid = this.valid();   /* and then check its validity */
      } while (!valid);         /* while hole sequence is invalid */
      this.state = 1;           /* reinitialize the path state */
    }                           /* for the end vertex traversal */
  }  /* next() */

  /*------------------------------------------------------------------*/
  /** Split a subproblem with the current path.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void split ()
  {                             /* --- split subproblem with path */
    int     i, k, d, e, n;      /* loop variables, buffers */
    Vertex  v, p, q, h;         /* buffers for vertices */
    boolean lft2rgt;            /* flag for a left to right edge */
    boolean vlp, vlq;           /* flags for rel. anchor position */

    n = this.verts.length;      /* get modulo for id computations */

    /* --- path cuts off the anchor --- */
    if (this.plen <= 0) {       /* if the path cuts off the anchor */
      v       = this.key[0];    /* get the subproblem anchor */
      p       = this.frst;      /* and the vertices at the start */
      q       = this.last;      /* and at the end of the path */
      vlp     = (v.compareTo(p) <= 0);
      vlq     = (v.compareTo(q) <= 0);
      lft2rgt = (p.compareTo(q) <= 0);

      /* -- build the left subproblem -- */
      if (vlp && vlq) {         /* if the old anchor is leftmost */
        i = 0;                  /* default: keep the old anchor */
        if (v.id < n) {         /* if anchor is a perimeter vertex */
          if      ((v.id +n +this.dir) % n == p.id) {
            if ( lft2rgt || (q.id < n)) i = -1; }
          else if ((v.id +n -this.dir) % n == q.id) {
            if (!lft2rgt || (p.id < n)) i = +1; }
        } }                     /* adapt the anchor if necessary */
      else if (lft2rgt) {       /* if path start is leftmost vertex */
        i = (vlq && (v.id < n) && ((v.id +n +this.dir) % n == p.id))
          ? -1 : +1; }          /* adapt the anchor if necessary */
      else {                    /* if path end is leftmost vertex */
        i = (vlp && (v.id < n) && ((v.id +n -this.dir) % n == q.id))
          ? +1 : -1;            /* adapt the anchor if necessary */
      }
      /* Note that the left subproblem is a triangle without holes    */
      /* (see the corresponding check in the method "valid" above).   */
      /* Thus, in principle, it does not matter which vertex is its   */
      /* anchor, because there will be no recursion anyway. However,  */
      /* for the lookup in the trie, the key should be unique. Hence, */
      /* as a general rule, the leftmost vertex is made the anchor.   */
      /* There are a few exceptions, though, which arise due to the   */
      /* fact that triangles can also result from a path starting at  */
      /* the anchor. These special cases are (*: perimeter edge):     */
      /*   v---q       q         q       v***q        v       q---p   */
      /*    * /       * \       / \       \ /        * \       * /    */
      /*     p       v---p     v***p       p        p---q       v     */
      /* The first four are handled in the case (vlp && vlq). In the  */
      /* first two the vertex opposite to the perimeter edge has to   */
      /* be made the anchor unconditionally, in the third and fourth  */
      /* case it has to be made the anchor only if it is a perimeter  */
      /* vertex. Otherwise v should be the anchor. In the other two   */
      /* cases the vertex opposite to the perimeter egdes must always */
      /* be the anchor. The checks identify the perimeter edge.       */

      this.lftcnt = this.lftall = 3; /* left subproblem is a triangle */
      if      (i == 0) {        /* if to keep the old anchor */
        this.lftkey[0] = v;     /* copy the old anchor first */
        this.lftkey[1] = p;     /* and complete the subproblem */
        this.lftkey[2] = q; }   /* by the path */
      else if (i >  0) {        /* if to make path start the anchor */
        this.lftkey[0] = p;     /* start the subproblem with the path */
        this.lftkey[1] = q;     /* and add the old anchor at the end */
        this.lftkey[2] = v; }
      else {                    /* if to make path end the anchor */
        this.lftkey[0] = q;     /* make path end the anchor, */
        this.lftkey[1] = v;     /* continue through old anchor */
        this.lftkey[2] = p;     /* back to the start of the path */
      }

      /* -- find anchor of right subproblem -- */
      d = 0;                    /* default: do not search new anchor */
      if      (vlp && vlq) {    /* if anchor is left  of the path */
        if (lft2rgt) { v = p; i = 1;          }
        else         { v = q; i = this.cnt-1; } }
        /* If the anchor is left of the path, the left of the two   */
        /* path vertices is the new anchor and no search is needed. */
      else if (vlp || vlq) {    /* if old anchor is in the middle, */
        d = i = 1; }            /* we has to search for a new anchor */
      else {                    /* if anchor is right of the path */
        if (lft2rgt) { v = q; i = this.cnt-1; }
        else         { v = p; i = 1;          }
        if (v.id < n) d = 1;    /* check whether a search is needed */
        /* If the anchor is right of the path, the right of the two */
        /* path vertices is, by default, the new anchor. However,   */
        /* if this default anchor is a perimeter vertex, one has    */
      } /* to search for a new anchor in the opposite direction.    */
      if (d == 0)               /* if the default anchor is valid, */
        d = (v == p) ? +1 : -1; /* set the flag for the copying */
      else {                    /* if to search for a new anchor */
        if (lft2rgt) { v = p; i = 1;          d = +1; }
        else         { v = q; i = this.cnt-1; d = -1; }
        if (v.id >= n) {        /* if the new anchor is a hole */
          while (true) {        /* anchor search loop */
            h = this.key[i += d];  /* traverse the key elements */
            if (h.id < n) {     /* if next vertex is on perimeter, */
              v  = h; break; }  /* it is the new anchor, so abort */
            if (h.compareTo(v) > 0) {
              i -= d; break; }  /* if v is the leftmost hole vertex, */
            v = h;              /* v is the new anchor, so abort, */
          }                     /* otherwise continue the search */
        }                       /* (result: v contains new anchor) */
      }
      e = i;                    /* note the key index of vertex v */
      this.rgtkey[0] = v;       /* store the new anchor and */
      this.rgtcnt    = 1;       /* initialize the vertex counters */
      this.rgtall    = 0;       /* (vertices in key and on perimeter) */
      k = this.cnt-1;           /* get index of last key element */

      /* -- subproblem key starts before the path edge -- */
      if (d > 0) {              /* if to start before the path edge */
        if (this.key[1] == null) { /* if start in perimeter section */
          d = this.key[2].id -p.id;
          if (this.dir < 0) d = -d;
          if (d <= 0) d += n;   /* get size of remaining section */
          if (d >  1) {         /* add perimeter section if necessary */
            this.rgtkey[this.rgtcnt++] = null; this.rgtall = d-2; }
        }
        while (++i < k)         /* copy key part after new anchor */
          this.rgtkey[this.rgtcnt++] = this.key[i];
        if ((this.key[k] == null)
        &&  (k > 1)) {          /* if end is in perimeter section */
          d = q.id -this.key[k-1].id;
          if (this.dir < 0) d = -d;
          if (d <= 0) d += n;   /* get size of remaining section */
          if (d >  1) {         /* add perimeter section if necessary */
            this.rgtkey[this.rgtcnt++] = null; this.rgtall = d-2; }
        }
        this.rgtkey[this.rgtcnt++] = q;
        for (i = 0; ++i < e; )  /* copy key part up to new anchor */
          this.rgtkey[this.rgtcnt++] = this.key[i];
        /* If the subproblem key starts before the path edge, it has */
        /* to be built from a part at the end of the old key (after  */
        /* the new anchor) followed by the path edge and the then    */
      } /* remaining part from the start of the old subproblem key.  */

      /* -- subproblem key starts after the path edge -- */
      else {                    /* if to start after the path edge */
        while (++i <= k)        /* copy key part after the new anchor */
          this.rgtkey[this.rgtcnt++] = this.key[i];
        this.rgtkey[this.rgtcnt++] = p;
        if (this.key[1] == null) { /* if start in perimeter section */
          d = this.key[2].id -p.id;
          if (this.dir < 0) d = -d;
          if (d <= 0) d += n;   /* get size of remaining section */
          if (d >  1) {         /* add perimeter section if necessary */
            this.rgtkey[this.rgtcnt++] = null; this.rgtall = d-2; }
        }
        for (i = 1; ++i < e; )  /* copy key part up to new anchor */
          this.rgtkey[this.rgtcnt++] = this.key[i];
        if ((this.key[k] == null)
        &&  (k > 1)) {          /* if end is in perimeter section */
          d = q.id -this.key[k-1].id;
          if (this.dir < 0) d = -d;
          if (d <= 0) d += n;   /* get size of remaining section */
          if (d >  1) {         /* add perimeter section if necessary */
            this.rgtkey[this.rgtcnt++] = null; this.rgtall = d-2; }
        }
        /* If the subproblem key starts after the path edge, it has   */
        /* to be built from a part at the start of the old subproblem */
        /* key, followed the path edge and then the remaining part    */
      } /* from the end of the old subproblem key.                    */
      this.rgtall += this.rgtcnt;
      return;                   /* subproblems have been built, */
    }                           /* so abort the function */

    /* --- path starts at the anchor --- */
    /* The general idea of the split is to traverse the elements of  */
    /* the current key up to the end of the path and then to follow  */
    /* the reversed path. This yields the right subproblem (right of */
    /* the splitting path). The left subproblem is constructed from  */
    /* the path (forward direction) and the remaining key elements.  */
    p = this.key[0];            /* get the subproblem anchor */
    q = this.path[this.plen-1]; /* and the end of the path */
    lft2rgt = (p.compareTo(q) <= 0);

    /* -- find split point in current key -- */
    this.rgtkey[0] = p;         /* always copy the subproblem anchor */
    d = e = 0;                  /* init. span and path end position */
    for (i = 0; true; ) {       /* traverse the key elements */
      v = this.key[++i];        /* copy all vertices in the key up to */
      if (v == q) break;        /* but excluding the end of the path */
      if (v != null) {          /* if not at the perimeter section */
        if ((v.id < n) && (e <= 0))
          e = 1; }              /* note the end of the ccw. path */
      else {                    /* if at the perimeter section */
        p = this.key[(i -this.dir) % this.cnt];
        e = this.key[(i +this.dir) % this.cnt].id -p.id;
        if (e <= 0) e += n;     /* (e is number of edges in section) */
        if (q.id < n) {         /* if path ends at perimeter vertex */
          d = (q.id +n -p.id) % n;
          if (this.dir < 0) d = e-d;
          if (d < e) break;     /* if end of path in section, abort */
        }                       /* (d states position in section) */
      }
      this.rgtkey[i] = v;       /* copy the current key vertex */
    }                           /* (partially build right subproblem) */

    /* -- build key of right subproblem -- */
    if      (!lft2rgt           /* if path leads to the left and */
    &&       (e <= 0)) {        /* ends on the counterclockwise path */
      System.arraycopy(this.rgtkey, 0, this.rgtkey, this.plen, i);
      this.rgtcnt = 0;          /* shift copied section to the end */
      for (k = this.plen; --k >= 0; )  /* start with reversed path */
        this.rgtkey[this.rgtcnt++] = this.path[k];
      this.rgtall = this.rgtcnt += i; }
    else if (!lft2rgt           /* if for a path from right to left */
    &&       (v == null)        /* there is no ccw path in the key, */
    &&       (i <= 1)) {        /* the subproblem key must start */
      this.rgtcnt = 0;          /* with the reversed path */
      for (k = this.plen; --k >= 0; )
        this.rgtkey[this.rgtcnt++] = this.path[k];
      this.rgtkey[this.rgtcnt++] = this.key[0];
      if (d <= 1)               /* if there is no perimeter section */
        this.rgtall = this.rgtcnt;  /* set total number of vertices */
      else {                    /* otherwise add perimeter section */
        this.rgtkey[this.rgtcnt++] = null;
        this.rgtall = this.rgtcnt +d-2;
      } }                       /* compute total number of vertices */
    else {                      /* if normal subproblem type */
      this.rgtcnt = i;          /* set the number of copied vertices */
      if (v != null)            /* if end not in perimeter section, */
        this.rgtall = (e > 2) ? e-2 : 0;    /* get the section span */
      else if (d <= 1)          /* if no vertex left in section, */
        this.rgtall = 0;        /* there are no additional vertices */
      else {                    /* add perimeter section to key */
        this.rgtall = d-2; this.rgtkey[this.rgtcnt++] = null; }
      for (k = this.plen; --k >= 0; )
        this.rgtkey[this.rgtcnt++] = this.path[k];
      this.rgtall += this.rgtcnt;      /* add the reversed path and */
    }                           /* compute total number of vertices */
    /* If the path leads from right to left and ends at a hole vertex */
    /* or at the end of the counterclockwise splitting path, the end  */
    /* of the path is the new anchor, otherwise the anchor is kept.   */

    /* -- build key of left subproblem -- */
    if      (!lft2rgt           /* if path leads to the left and ends */
    &&       (e > 0)            /* on the clockwise path (neither on */
    &&       (v != null)) {     /* ccw path nor on perimeter section) */
      this.lftcnt = 0;          /* initialize vertex counter */
      for (k = i; k < this.cnt; k++)  /* start with rest of curr. key */
        this.lftkey[this.lftcnt++] = this.key[k];
      this.lftkey[this.lftcnt++] = this.key[0];
      i = this.plen-1;          /* add the path (forward) with the */
      for (k = 0; k < i; k++)   /* exception of the last path vertex */
        this.lftkey[this.lftcnt++] = this.path[k];
      this.lftall = this.lftcnt; }
    else if (!lft2rgt           /* if for a path from right to left */
    &&       (v == null)        /* there is no cw path in the key, */
    &&       (i >= this.cnt-1)){/* the subproblem key must start */
      this.lftkey[0] = q;       /* start with the end of the path */
      this.lftcnt    = 1;       /* initialize the vertex counter */
      if (d+1 >= e)             /* if no vertex left in section, */
        e = 0;                  /* there are no additional vertices */
      else {                    /* add perimeter section to key */
        e -= d+2; this.lftkey[this.lftcnt++] = null; }
      this.lftkey[this.lftcnt++] = this.key[0];
      i = this.plen-1;          /* add the path (forward) with the */
      for (k = 0; k < i; k++)   /* exception of the last path vertex */
        this.lftkey[this.lftcnt++] = this.path[k];
      this.lftall = this.lftcnt +e; }
    else {                      /* if normal subproblem type */
      this.lftall = 0;          /* initialize vertex counter */
      this.lftkey[0] = this.key[0];
      for (k = this.plen; --k >= 0; )
        this.lftkey[k+1] = this.path[k];
      this.lftcnt = this.plen+1;/* start with the path (forward) */
      if ((v != null)           /* if end not in perimeter section */
      ||  (d+1 >= e))           /* or no vertex left in section, */
        e = 0;                  /* there are no additional vertices */
      else {                    /* add perimeter section to key */
        e -= d+2; this.lftkey[this.lftcnt++] = null; }
      while (++i < this.cnt) {  /* add the rest of the current key */
        this.lftkey[this.lftcnt++] = v = this.key[i];
        if (v != null) continue;/* if not at perimeter section, skip */
        p = this.key[(i -this.dir) % this.cnt];
        e = this.key[(i +this.dir) % this.cnt].id -p.id;
        if (e <= 0) e += n;     /* (e: number of edges in section) */
        e -= 2;                 /* subtract vertices that are */
      }                         /* already on the subproblem key */
      this.lftall = this.lftcnt +e;
    }                           /* compute total number of vertices */
    /* If the path leads from right to left and ends at a hole vertex */
    /* or at the end of the clockwise splitting path, the end of the  */
    /* path is the new anchor, otherwise the old anchor is kept.      */
  }  /* split() */

  /*------------------------------------------------------------------*/
  /** Get the current triangle.
   *  @return always <code>null</code>
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Vertex[] triangle ()
  { return null; }

  /*------------------------------------------------------------------*/
  /** Compute the weight of the current split.
   *  @return the weight of the current split
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected double weight ()
  {                             /* --- compute weight of split */
    if (this.plen == 0)         /* if path cuts off the anchor */
      return -this.wgts[this.frst.id][this.last.id];
    double wgt = -this.wgts[this.key[0].id][this.path[0].id];
    for (int i = this.plen; --i > 0; )
      wgt -= this.wgts[this.path[i].id][this.path[i-1].id];
    return wgt;                 /* sum the edge lengths and */
  }  /* weight() */             /* return the computed path length */

  /*------------------------------------------------------------------*/
  /** Get the edges of the current path.
   *  @return the edges of the current path as an array
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Edge[] edges ()
  {                             /* --- get edges of current path */
    int    i;                   /* loop variable */
    Edge[] edges;               /* edges of current path */

    if (this.plen <= 0) {       /* if path cuts off the anchor, */
      edges    = new Edge[1];   /* create and return the only edge */
      edges[0] = new Edge(this.frst, this.last); }
    else {                      /* if path starts at the anchor */
      for (edges = new Edge[i = this.plen]; --i > 0; )
        edges[i] = new Edge(this.path[i-1], this.path[i]);
      edges[0] = new Edge(this.key[0], this.path[0]);
    }                           /* create and return the edges */
    return edges;               /* on the splitting path */
  }  /* edges() */

  /*------------------------------------------------------------------*/
  /** Store the current path as best.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void store ()
  {                             /* --- store current path as best */
    this.bestlen = this.plen;   /* store the path length */
    if (this.plen > 0)          /* if the path is not empty, copy it */
      System.arraycopy(this.path, 0, this.best, 0, this.plen);
  }  /* store() */

  /*------------------------------------------------------------------*/
  /** Retrieve the best path.
   *  @return the best path
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Object retrieve ()
  {                             /* --- retrieve best path */
    if (this.bestlen <= 0)      /* if path cuts off the anchor, */
      return Path.cut;          /* return fixed dummy empty path */
    if (this.bestlen <= 1)      /* single vertices are used directly */
      return ((Vertex[])this.best)[0];
    Vertex[] p = new Vertex[this.bestlen];
    System.arraycopy(this.best, 0, p, 0, this.bestlen);
    return p;                   /* copy the best path into a */
  }  /* retrieve() */           /* fitting vector and return it */

  /*------------------------------------------------------------------*/
  /** Distribute the hole vertices.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void distribute ()
  {                             /* --- distribute the hole vertices */
    int    i, k, d;             /* loop variables, increment */
    int    lft, rgt;            /* number of edges of subpointgons */
    Vertex h;                   /* to traverse the hole vertices */

    this.lftholecnt = 0;        /* initialize the hole counters */
    this.rgtholecnt = 0;        /* and check the number of holes */
    if (this.holecnt <= 0) return;
    lft = this.size(this.lftkey, this.lftcnt);
    rgt = this.size(this.rgtkey, this.rgtcnt);
    if (lft +rgt <= 0) return;  /* check for empty triangles */

    /* --- get the path traversal parameters --- */
    if (this.pdir < 0) { d = -1; k = this.plen-2; }
    else               { d = +1; k = 0; this.path[this.plen] = null; }

    /* --- if not to distribute hole vertices --- */
    if (!this.dist) {           /* if not to distribute holes */
      if ((this.plen <= 0)      /* if path cuts off the anchor */
      &&  (rgt       <= 3)) {   /* and path splits a quadrangle */
        for (i = 0; i < this.holecnt; i++) {
          h = this.holes[i];    /* traverse the hole vertices */
          if (this.inside(h, this.rgtkey, this.rgtcnt))
            this.rgtholes[this.rgtholecnt++] = h;
        } return;               /* check against right triangle */
      }                         /* and abort the function */
      if ((lft <= 3)            /* if both subproblems are */
      &&  (rgt <= 3)) {         /* either triangles or empty */
        for (i = 0; i < this.holecnt; i++) {
          h = this.holes[i];    /* traverse the hole vertices */
          if      ((lft >= 3)   /* check left subproblem */
          &&       this.inside(h, this.lftkey, this.lftcnt))
            this.lftholes[this.lftholecnt++] = h;
          else if ((rgt >= 3)   /* check right subproblem */
          &&       this.inside(h, this.rgtkey, this.rgtcnt))
            this.rgtholes[this.rgtholecnt++] = h;
        } return;               /* check against both triangles */
      }                         /* and abort the function */
      if ((lft > 3)             /* if neither of the subproblems */
      &&  (rgt > 3)) {          /* is a triangle */
        for (i = 0; i < this.holecnt; i++) {
          h = this.holes[i];    /* traverse the hole vertices */
          if (h == this.path[k]) {   /* skip holes on the path */
            k += d; if (k < 0) k = 0; continue; }
          this.lftholes[this.lftholecnt++] = h;
          this.rgtholes[this.rgtholecnt++] = h;
        } return;               /* store all holes (except those */
      }                         /* that are on the path) for both */
    }                           /* subproblems */
    /* The above distribution of hole vertices takes time linear in  */
    /* the number of holes, since if a check is carried out at all   */
    /* whether a hole is inside a subproblem, it is a triangle.      */
    /* In cases that are not handled in the above code, but below,   */
    /* the function "inside" is also carried out only for triangles. */

    /* --- if to distribute hole vertices --- */
    if (this.plen <= 0) {       /* if path cuts off the anchor */
      for (i = 0; i < this.holecnt; i++)
        this.rgtholes[this.rgtholecnt++] = this.holes[i]; }
      /* Since the triangle cut off by a path cutting off the anchor  */
      /* must not contain any holes (see check in is function valid), */
      /* all holes not on the path must be in the right subproblem.   */
    else if ((lft < 3)          /* if one of the subproblems */
    ||       (rgt < 3)) {       /* is empty (does not exist) */
      for (i = 0; i < this.holecnt; i++) {
        h = this.holes[i];      /* traverse the hole vertices, */
        if (h == this.path[k]){ /* but skip holes on the path */
          k += d; if (k < 0) k = 0; continue; }
        if (lft < 3) this.rgtholes[this.rgtholecnt++] = h;
        else         this.lftholes[this.lftholecnt++] = h;
      } }                       /* store hole in corresp. vector */
    else if (lft < rgt) {       /* if left subprob. has fewer edges */
      for (i = 0; i < this.holecnt; i++) {
        h = this.holes[i];      /* traverse the hole vertices, */
        if (h == this.path[k]){ /* but skip holes on the path */
          k += d; if (k < 0) k = 0; continue; }
        if (this.inside(h, this.lftkey, this.lftcnt))
             this.lftholes[this.lftholecnt++] = h;
        else this.rgtholes[this.rgtholecnt++] = h;
      } }                       /* store hole in corresp. vector */
    else {                      /* if right subprob. has fewer edges */
      for (i = 0; i < this.holecnt; i++) {
        h = this.holes[i];      /* traverse the hole vertices, */
        if (h == this.path[k]){ /* but skip holes on the path */
          k += d; if (k < 0) k = 0; continue; }
        if (this.inside(h, this.rgtkey, this.rgtcnt))
             this.rgtholes[this.rgtholecnt++] = h;
        else this.lftholes[this.lftholecnt++] = h;
      }                         /* check which subproblem hole is in */
    }                           /* and store hole in corresp. vector */
    /* In order to optimize speed, the "inside" test is carried out */
    /* w.r.t. the subproblem with fewer vertices on the perimeter.  */
  }  /* distribute() */

}  /* class Path */
