/*----------------------------------------------------------------------
  File    : Triangle.java
  Contents: split a (sub-)pointgon by cutting out triangles
  Author  : Christian Borgelt
  History : 2005.04.21 file created from file Splitter.java
            2005.04.26 function "cut" completed, "inside" adapted
            2005.04.27 function "distribute" completed
            2005.05.12 construction of cut out triangle added
            2005.06.22 perimeter section represented explicitly
            2005.06.23 check for path vertices inside triangle added
            2005.06.25 bug in function edgecnt fixed
            2005.12.15 use of more precomputed information added
            2005.12.28 changed to extension of abstract class
            2006.01.04 parameter mwt of init method moved to constructor
            2006.01.21 hole distribution slightly improved
            2006.01.27 triangle moved to abstract class Splitter
            2006.02.05 some functions moved to abstract class Splitter
            2006.02.06 function for getting the subproblem size improved
----------------------------------------------------------------------*/
package pointgon;

/*--------------------------------------------------------------------*/
/** Class for splitting a (sub-)pointgon into (sub-)pointgons.
 *  @since  2006.01.27 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class Triangle extends Splitter {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00010000;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** tip (third corner of triangle) */
  private Vertex tip;

  /*------------------------------------------------------------------*/
  /** Get next tip of a triangle.
   *  @return whether a another tip of a traingle was found
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean nextTip ()
  {                             /* --- get next tip of triangle */
    int    n;                   /* number of perimeter vertices */
    Vertex v;                   /* buffer for next vertex */

    if (this.state <= 0) {      /* if possible, get the next hole */
      this.tip = this.holes[-this.state++]; return true; }
    if (this.state == 1)        /* after holes get second key vertex */
      this.tip = this.key[this.state++];
    if (this.state < this.cnt){ /* if before the end of the key, */
      v = this.key[this.state++];     /* get the next key vertex */
      if (v != null) { this.tip = v; return true; }
      this.state++;             /* check for a perimeter section */
    }                           /* and note it in the state */
    if (this.state > this.cnt){ /* if in the perimeter section, */
      n = this.verts.length;    /* get the next perimeter vertex */
      this.tip = this.verts[(this.tip.id +n +this.dir) % n];
      return (this.tip != this.key[0]);
    }                           /* check for wrap around */
    return false;               /* no next tip/triangle corner */
  }  /* nextTip() */

  /*------------------------------------------------------------------*/
  /** Check wether the current triangle is valid.
   *  @return whether the triangle is valid
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean valid ()
  {                             /* -- check whether triangle is valid */
    int    i, n;                /* loop variable, number of vertices */
    Vertex a, b, c;             /* corners of the triangle */
    Vertex p, q;                /* to traverse the key vertices */

    /* --- check for a proper triangle --- */
    a = this.key[0];            /* get the three corners */
    b = this.key[1];            /* of the current triangle */
    c = this.tip;               /* (base and tip) */
    if (!this.valid(a, b, c))   /* if the triangle is invalid, */
      return false;             /* abort the function */

    /* --- check for intersection with path --- */
    q = this.key[2];            /* get the second path edge */
    if (q == null) return true; /* if there is none, abort function */
    if (q != c) {               /* if third vertex is not the tip */
      if ((this.trtab == null)  /* if there is no precomputed info.: */
      &&  q.isInside(a, b, c))  /* the third vertex must not */
        return false;           /* be inside the triangle */
      if (a.isectsX(c, b, q))   /* check intersection with side (a,c) */
        return false;           /* (left side of the triangle) */
    }
    for (i = 2; ++i < this.cnt; ) {
      p = q; q = this.key[i];   /* get the next edge of the path */
      if (q == null) break;     /* if there is none, abort the loop */
      if ((p == c)              /* if there is no precomputed info.: */
      &&  (this.trtab == null)  /* if the vertex after the tip */
      &&  q.isInside(a, b, c))  /* is inside the cut out triangle, */
        return false;           /* the triangle is invalid */
      if ((p != c) && (q != c)  /* skip edges incident to the tip */
      &&  (a.isectsX(c, p, q) || b.isectsX(c, p, q)))
        return false;           /* check intersection of the edge */
    }                           /* with either side of the triangle */
    if ((q != c) && (q != null) /* if there is a wrap around edge, */
    &&  a.isectsX(q, b, c))     /* check intersection with side (b,c) */
      return false;             /* (right side of the triangle) */
    return true;                /* return 'triangle is valid' */
  }  /* valid() */

  /*------------------------------------------------------------------*/
  /** Check whether a hole is inside a subproblem.
   *  @param  hole  the hole to check
   *  @param  key   the key  of the subproblem
   *  @param  cnt   the size of the subproblem
   *  @return whether the hole is insode the subproblem
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean inside (Vertex hole, Vertex[] key, int cnt)
  {                             /* --- check whether a hole is inside */
    int    i, r, n, c = 0;      /* loop variables, buffer, counter */
    Vertex p, q, v;             /* perimeter vertices of subproblem */

    n = (key[cnt-1] == null) ? cnt-1 : cnt;
    p = v = key[0];             /* traverse the vertices on the */
    for (i = 0; ++i < n; p = q){/* perimeter of the given subproblem */
      q = key[i];               /* check whether a horiz. line cuts */
      r = hole.horzLineCuts(p, q);     /* the next edge on the path */
      if (r == 0) return false; /* if vertex is on the edge, abort */
      if (r >  0) c++;          /* if horz. line cuts to the right */
    }                           /* of the vertex, count the edge */
    n = this.verts.length;      /* get modulo for id computations */
    q = (i >= cnt) ? v : this.verts[(p.id +n +this.dir) % n];
    while (true) {              /* traverse the perimeter section */
      r = hole.horzLineCuts(p, q); /* check whether horiz. line cuts */
      if (r == 0) return false; /* if vertex is on the edge, abort */
      if (r >  0) c++;          /* if horz. line cuts to the right */
      if (q == v) break;        /* check whether horiz. line cuts */
      p = q; q = this.verts[(p.id +n +this.dir) % n];
    }                           /* get the next perimeter vertex */
    /* The idea of the above check is to shoot a straight line from   */
    /* the given point to the right and to count how many times this  */
    /* line cuts the perimeter. If the number is even, the point lies */
    /* outside the perimeter, if it is odd, the point lies inside.    */
    return (c & 1) != 0;        /* check the number of intersections */
  }  /* inside() */             /* (odd: inside, even: outside) */

  /*------------------------------------------------------------------*/
  /** Get the full size of a subproblem.
   *  @param  key   the key of the subproblem
   *  @param  cnt   the size of the subproblem
   *  @return the full size of a subproblem
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private int size (Vertex[] key, int cnt)
  {                             /* --- get full size of subproblem */
    int    k, n;                /* full size, number of vertices */
    Vertex p, q;                /* two corners of the triangle */

    if (cnt < 3) return 0;      /* check for an empty subproblem */
    if (key[--cnt] != null)     /* if there is no perimeter section, */
      return cnt+1;             /* return the key size */
    k = key[0].id -key[cnt-1].id;
    if (this.dir < 0) k = -k;   /* compute edges in perim. section */
    if (k <= 0) k += this.verts.length;
    k += cnt-1;                 /* compute the total number of edges */
    if ((k > 3) || (this.trtab == null))
      return k;                 /* check size and precomputed info. */
    p = key[1]; q = key[2];     /* get the two triangle corners */
    if (q == null) {            /* if there is a perimeter section */
      n = this.verts.length;    /* get modulo for id computation */
      q = this.verts[(p.id +n +this.dir) % n];
    }                           /* get vertex in perimeter section */
    return (this.trtab[key[0].id][p.id][q.id] != 0)
         ? 0 : k;               /* check for an empty triangle and */
  }  /* size() */               /* return the corr. subproblem size */

  /*------------------------------------------------------------------*/
  /** Create a triangle splitter.
   *  @param  mwt   the minimum weight triangulation object
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Triangle (MWT mwt)
  {                             /* --- create a subproblem splitter */
    int n = mwt.holes.length;   /* number of holes */
    this.dist     = ((mwt.mode & MWT.DISTHOLES) != 0);
    this.verts    = mwt.verts;  /* store perimeter vertices */
    this.dir      = mwt.dir;    /* and their order */
    this.wgts     = mwt.wgts;   /* note the precomputed information */
    this.trtab    = mwt.trtab;  /* and create the vertex vectors */
    this.triangle = new Vertex[3];
    this.rgtkey   = new Vertex[n +3];
    this.lftkey   = new Vertex[n +3];
    this.rgtholes = new Vertex[n];
    this.lftholes = new Vertex[n];
  }  /* Triangle() */

  /*------------------------------------------------------------------*/
  /** Initialize the path splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  holes    the holes of the subproblem
   *  @param  holecnt  the number of holes of the subproblem
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void init (Vertex[] key,   int cnt,
                       Vertex[] holes, int holecnt)
  {                             /* --- init. a subproblem splitter */
    this.key     = key;         /* note the key */
    this.cnt     = cnt;         /* of the subproblem */
    this.holes   = holes;       /* note the holes that are (possibly) */
    this.holecnt = holecnt;     /* inside this subproblem */
    this.state   = 1-holecnt;   /* initialize the path state */
    this.best    = null;        /* and the best triangle tip */
  }  /* init() */

  /*------------------------------------------------------------------*/
  /** Initialize a path splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  spobj    the splitter object
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void init (Vertex[] key, int cnt, Object spobj)
  {                             /* --- init. a subproblem splitter */
    this.key = key;             /* note the key */
    this.cnt = cnt;             /* of the subproblem */
    this.tip = (Vertex)spobj;   /* and the triangle tip */
  }  /* init() */

  /*------------------------------------------------------------------*/
  /** Create the next subproblem split.
   *  @return whether a new split was created
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean next ()
  {                             /* --- create next subproblem split */
    do {                        /* triangle search loop */
      if (!this.nextTip())      /* if there is no next triangle tip, */
        return false;           /* abort the function with failure */
    } while (!this.valid());    /* repeat until the triangle is valid */
    return true;                /* return 'next triangle found' */
  }  /* next() */

  /*------------------------------------------------------------------*/
  /** Split a subproblem with the current triangle.
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void split ()
  {                             /* --- cut triangle out of subproblem */
    int    i, k, n;             /* loop variables, number of vertices */
    int    lft;                 /* key index of leftmost hole vertex */
    Vertex v, end;              /* to traverse the vertices, buffer */

    this.triangle[0] = null;    /* initialize the triangle and */
    end = this.key[this.cnt-1]; /* get the last vertex on the path */
    if (end == null) end = this.key[this.cnt-2];
    n   = this.verts.length;    /* get number of perimeter vertices */
    if ((this.tip.id < n)       /* if the tip is on the perimeter, */
    &&  (this.tip   != end)     /* but it is not the end of the path, */
    &&  (end.id      < n)) {    /* and the path ends on the perimeter */
      v = this.verts[(this.tip.id +n +this.dir) % n];
      if (this.key[0] == v) {   /* if tip is cw neighbor of base, */
        this.lftcnt = 0;        /* there is no left subproblem */
        this.triangle[0] = this.key[0];
        this.triangle[1] = this.key[1];
        this.triangle[2] = this.tip; }
      else {                    /* if part left of triangle exists */
        this.lftkey[0] = this.key[0];
        this.lftkey[1] = this.tip;
        this.lftkey[2] = null;  /* store the two vertices on the path */
        this.lftcnt    = 3;     /* and the perimeter section, */
      }                         /* then set the vertex counter */
      v = this.verts[(this.tip.id +n -this.dir) % n];
      if (this.key[1] == v) {   /* if tip is ccw neighbor of base, */
        this.rgtcnt = 0;        /* there is no right subproblem */
        this.triangle[0] = this.tip;
        this.triangle[1] = this.key[0];
        this.triangle[2] = this.key[1]; }
      else {                    /* if part right of triangle exists */
        this.rgtcnt = (v == end) ? this.cnt-1 : this.cnt;
        System.arraycopy(this.key, 1, this.rgtkey, 1, this.rgtcnt-1);
        this.rgtkey[0] = this.tip; /* copy the current key and */
      } }                          /* replace the first vertex */
    else {                      /* if the tip is a hole vertex */
      this.lftkey[0] = this.key[0]; /* or the last path vertex */
      this.lftkey[1] = v = this.tip;
      this.lftkey[2] = this.key[1];
      for (k = 2; k < this.cnt; ) {
        v = this.key[k];        /* traverse the key vertices, */
        if (v == this.tip) break;      /* but stop at the tip */
        this.lftkey[++k] = v;   /* copy the key vertex in order */
      }                         /* to build the left subproblem */
      if      (k >= this.cnt) { /* if the tip is not on the path, */
        this.lftcnt = k+1;      /* all vertices are in the left key */
        this.rgtcnt = 0; }      /* and there is no right subproblem */
      else if (k <= 2) {        /* if tip is the vertex ccw of base */
        for (i = 2; ++k < this.cnt; )
          this.lftkey[i++] = this.key[k];
        this.lftcnt = i;        /* build key of the left subproblem */
        this.rgtcnt = 0; }      /* and clear the right subproblem */
      else {                    /* if tip is somewhere on the path */
        lft = k;                /* default: tip is leftmost hole */
        for (i = 0; ++i < k; )  /* find the leftmost hole vertex */
          if (this.key[i].compareTo(v) < 0)
            v = this.key[lft = i];
        i = k -lft +1;          /* build the right subproblem key */
        System.arraycopy(this.key, lft, this.rgtkey, 0, i);
        System.arraycopy(this.key, 1,   this.rgtkey, i, lft-1);
        this.rgtcnt = k;        /* copy the two sections (ccw and cw) */
        for (i = 2; ++k < this.cnt; )
          this.lftkey[i++] = this.key[k];
        this.lftcnt = (i > 2) ? i : 0;
      }                         /* build key of left subproblem */
    }                           /* and set the vertex counter */
    if (this.triangle[0] != null)
      return;                   /* check for constructed triangle */
    if ((this.tip == end)       /* if a one edge perimeter section */
    &&  (this.tip.id < n)       /* is covered by a triangle side */
    &&  (this.key[this.cnt-1] != null)) {
      this.triangle[0] = this.key[0];  /* keep standard order */
      this.triangle[1] = this.key[1];
      this.triangle[2] = this.tip; }
    else if ((this.tip.compareTo(this.key[0]) < 0)
    &&       (this.tip.compareTo(this.key[1]) < 0)) {
      this.triangle[0] = this.tip;     /* tip is leftmost corner */
      this.triangle[1] = this.key[0];
      this.triangle[2] = this.key[1]; }
    else if (this.key[0].compareTo(this.key[1]) <  0) {
      this.triangle[0] = this.key[0];  /* key[0] is leftmost corner */
      this.triangle[1] = this.key[1];
      this.triangle[2] = this.tip; }
    else { /* (this.key[0].compareTo(this.key[1]) >= 0) */
      this.triangle[0] = this.key[1];  /* key[1] is leftmost corner */
      this.triangle[1] = this.tip;
      this.triangle[2] = this.key[0];
    }                           /* start triangle at leftmost corner */
  }  /* split() */

  /*------------------------------------------------------------------*/
  /** Get the current triangle.
   *  @return the current triangle
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Vertex[] triangle ()
  { return this.triangle; }

  /*------------------------------------------------------------------*/
  /** Compute the weight of the current split.
   *  @return the weight of the current split
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected double weight ()
  {                             /* --- get weight of triangle cut */
    double wgt = this.wgts[this.key[0].id][this.key[1].id];
    if (this.lftcnt <= 0)       /* add left  edge of the triangle */
      wgt += this.wgts[this.key[0].id][this.tip.id];
    if ((this.rgtcnt <= 0)      /* add right edge of the triangle */
    &&  (this.lftcnt <= this.cnt))
      wgt += this.wgts[this.key[1].id][this.tip.id];
    return wgt;                 /* return weight of edges that are */
  }  /* weight() */             /* not part of the subproblems */

  /*------------------------------------------------------------------*/
  /** Get the edges of the current split.
   *  @return the edges of the current split as an array
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Edge[] edges ()
  {                             /* --- get edges of subproblem split */
    int    n;                   /* loop variable */
    Edge[] edges;               /* edges of current split */

    n = (((this.lftcnt > 0) && (this.rgtcnt > 0))
       || (this.lftcnt > this.cnt)) ? 2 : 1;
    edges = new Edge[n];        /* create properly sized edge vector */
    if  (this.lftcnt > 0)       /* edge cutting off left  subproblem */
      edges[--n] = new Edge(this.key[0], this.tip);
    if((this.rgtcnt > 0)        /* edge cutting off right subproblem */
    || (this.lftcnt > this.cnt))/* or right side of triangle to hole */
      edges[--n] = new Edge(this.key[1], this.tip);
    return edges;               /* create and return the edges */
  }  /* edges() */              /* of the best separating triangle */

  /*------------------------------------------------------------------*/
  /** Store the current split as best.
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void store ()
  { this.best = this.tip; }     /* --- store current tip as best */

  /*------------------------------------------------------------------*/
  /** Retrieve the best split.
   *  @return the best split
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Object retrieve ()
  { return this.best; }         /* --- retrieve best triangle tip */

  /*------------------------------------------------------------------*/
  /** Distribute the hole vertices.
   *  @since  2005.04.21 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void distribute ()
  {                             /* --- distribute the hole vertices */
    int    i;                   /* loop variable */
    int    lft, rgt;            /* number of edges of subpointgons */
    Vertex h;                   /* to traverse the hole vertices */

    this.lftholecnt = 0;        /* initialize the hole counters */
    this.rgtholecnt = 0;        /* and check the number of holes */
    if (this.holecnt <= 0) return;   /* then get number of edges */
    lft = this.size(this.lftkey, this.lftcnt);
    rgt = this.size(this.rgtkey, this.rgtcnt);
    if (lft +rgt <= 0) return;  /* check for empty triangles */

    /* --- if not to distribute hole vertices --- */
    if (!this.dist) {           /* if not to distribute holes */
      if ((lft <= 3)            /* if both subproblems are */
      &&  (rgt <= 3)) {         /* either triangles or empty */
        for (i = 0; i < this.holecnt; i++) {
          h = this.holes[i];    /* traverse the hole vertices, */
          if (h == this.tip) continue;     /* but skip the tip */
          if      ((lft >= 3)   /* check left subproblem */
          &&       this.inside(h, this.lftkey, this.lftcnt))
            this.lftholes[this.lftholecnt++] = h;
          else if ((rgt >= 3)   /* check right subproblem */
          &&       this.inside(h, this.rgtkey, this.rgtcnt))
            this.rgtholes[this.rgtholecnt++] = h;
        } return;               /* check against both triangles */
      }                         /* and abort the function */
      if ((lft > 3)             /* if neither subproblem */
      &&  (rgt > 3)) {          /* is a triangle */
        for (i = 0; i < this.holecnt; i++) {
          h = this.holes[i];    /* traverse the hole vertices */
          this.lftholes[this.lftholecnt++] = h;
          this.rgtholes[this.rgtholecnt++] = h;
        } return;               /* store all hole vertices */
      }                         /* for both subproblems */
    }                           /* (distribute only for triangles) */

    /* --- if to distribute hole vertices --- */
    if ((lft < 3)               /* if one of the subproblem */
    ||  (rgt < 3)) {            /* is empty (does not exist) */
      for (i = 0; i < this.holecnt; i++) {
        h = this.holes[i];      /* traverse the hole vertices, */
        if (h == this.tip) continue;       /* but skip the tip */
        if (lft < 3) this.rgtholes[this.rgtholecnt++] = h;
        else         this.lftholes[this.lftholecnt++] = h;
      } }                       /* store hole in corresp. vector */
    else if (lft < rgt) {       /* if left subprob. has fewer edges */
      for (i = 0; i < this.holecnt; i++) {
        h = this.holes[i];      /* traverse the hole vertices */
        if (this.inside(h, this.lftkey, this.lftcnt))
             this.lftholes[this.lftholecnt++] = h;
        else this.rgtholes[this.rgtholecnt++] = h;
      } }                       /* store hole in corresp. vector */
    else {                      /* if right subprob. has fewer edges */
      for (i = 0; i < this.holecnt; i++) {
        h = this.holes[i];      /* traverse the hole vertices */
        if (this.inside(h, this.rgtkey, this.rgtcnt))
             this.rgtholes[this.rgtholecnt++] = h;
        else this.lftholes[this.lftholecnt++] = h;
      }                         /* check which subproblem hole is in */
    }                           /* and store hole in corresp. vector */
  }  /* distribute() */

}  /* class Triangle */
